
require "util"
evo = {}
evo.do_this_once = true
global.rocket_launched = global.rocket_launched or false

local function build_tree(tree, tech)
	if game.forces["player"].technologies[tech].prerequisites then
		for i, prereq in pairs(game.forces["player"].technologies[tech].prerequisites) do
			local add_it = true
			for j, listed_tech in pairs(tree) do
				if prereq.name == listed_tech then add_it = false end
			end
			if add_it == true then
				table.insert(tree, prereq.name)
				log("inserting " .. prereq.name)
				build_tree (tree, prereq.name)
			end
		end
	end
end


function evo.modify_evolution()
	if global.critical_research_progress == nil then evo.set_critical_research_progress() end
	if evo.do_this_once == true then
		log ("Building Tech Tree")
		evo.leveling_techs = {{"rocket-silo"}}
		build_tree(evo.leveling_techs, "rocket-silo")
		evo.do_this_once = false	
	end
	if (global.last_evolution == 0 or global.last_evolution == nil) then global.last_evolution = game.forces["enemy"].evolution_factor end
	if global.rocket_launched == nil then -- regulate evolution based on inhibitor toxin and research progress
		local another_delta = (game.forces["enemy"].evolution_factor - global.last_evolution)
		if (global.evolution_suppression >.5) then -- slowing down evolution progress, regardless of what evolution is
			game.forces["enemy"].evolution_factor = game.forces["enemy"].evolution_factor - (.33*another_delta)
			global.toxin_levels = global.toxin_levels*.999 -- slowing down evolution costs toxin level
		end
		local delta = game.forces["enemy"].evolution_factor - global.critical_research_progress 
		if global.evolution_suppression >= 1 then 
			if delta >= 0 then -- completely halt evolution progress, if it's higher than the research progress
				game.forces["enemy"].evolution_factor = global.last_evolution
			end
			if delta < 0 then -- slow down evolution progress even more, but not stopped completely, if it's lower than the research progress
				game.forces["enemy"].evolution_factor = game.forces["enemy"].evolution_factor - (.33*another_delta)
			end
			global.toxin_levels = global.toxin_levels*.999 -- suppressing evolution at high toxin levens costs more toxin level
		end
	end
	global.last_evolution = game.forces["enemy"].evolution_factor 
end
	

	
function evo_onTick(event) -- set up intervals and run routines
	if (event.tick % 600 == 0) then 
		evo.modify_evolution() 		
		evo.set_critical_research_progress()
	end
end


function evo_onInit() --Set up the global tables that are persisted between loads
	global = global or {}
	global.last_evolution = global.last_evolution or 0
	global.critical_research_progress = global.critical_research_progress or 0
	global.rocket_launched = global.rocket_launched or false
end

function evo_onRocketLaunched(event)
	global.rocket_launched = true
end

function evo.set_critical_research_progress() 
	if evo.do_this_once == true then
		evo.leveling_techs = {{"rocket-silo"}}
		build_tree(evo.leveling_techs, "rocket-silo")
		evo.do_this_once = false	
	end
	local critical_techs_researched = 0
	for j, tech in pairs(game.forces["player"].technologies) do
		if tech.researched == true then
			for k=1, #evo.leveling_techs do
				if tech.name == evo.leveling_techs[k] then 
					critical_techs_researched = critical_techs_researched + 1 
				end				
			end
		end
	end
	local thermostat = (critical_techs_researched / #evo.leveling_techs)
	global.critical_research_progress = global.critical_research_progress or 0
	if global.critical_research_progress < thermostat then
		global.critical_research_progress = global.critical_research_progress + .000694 -- At this rate, it takes 4 hours to move the needle from 0 to 1.
	end 
end


