require "util"
weartear.interval = .1
weartear.aging_age = settings.global.aging_age.value / weartear.interval 
weartear.old_age = settings.global.old_age.value / weartear.interval 
weartear.dilapidated_age = settings.global.dilapidated_age.value / weartear.interval 
weartear.death_age = settings.global.death_age.value / weartear.interval 
weartear.interval = weartear.interval * 3600
weartear.idle_age = settings.global.idle_age.value
weartear.notifications = settings.global.notifications.value
weartear.emit_smoke = settings.global.smoke.value
weartear.mining_drills_age = settings.global.mining_drills_age.value
weartear.crafting_machines_age = settings.global.crafting_machines_age.value
weartear.steam_engines_age = settings.global.steam_engines_age.value
weartear.solar_panels_age = settings.global.solar_panels_age.value
weartear.furnaces_age = settings.global.furnaces_age.value
weartear.explosive_endings = settings.global.explosive_endings.value
weartear.when_to_replace = settings.global.when_to_replace.value
weartear.last_calculated_pollution = 0
weartear.refill_tick = {}
weartear.do_this_once = true
weartear.clean_schedule = 0
weartear.max_energy_production = {}
weartear.flavor_size = 37 -- this is how many flavor text varieties are in the localization file


function weartear.flavor_text() -- catalogue of flavor text for notifications.
	local text = {" has blown a gasket.", " is running like crap.", " has a faulty valve connection.", " needs a belt replacement.", " has burned out relay boards.", " is leaking fluids.", " needs a new hood switch.", " has a broken gearbox assembly.", " has a bad clamp cylinder.", " needs new fluids.", " is losing hydraulic pressure.", " has a faulty sensor.", " has a leaky hose.", " needs a new choke valve.", " has rusted out springs.", " needs a new coil spring.", " has a cracked belt housing.", " has rusted out cotter pins.", " could use some adhsesive tape.", " is leaking all over the place.", " is overheating.", " has broken gears.", " has a burned out fan clutch.", " has blown a fuse.", " has flashing indicator lights.", " needs a tune-up.", " has cracked slip-rings.", " has given up the ghost.", " has a bad water chip.", " needs a 12-point inspection.", " is making a rattling sound.", " has frayed cables.", " needs a new flywheel.", " is not operating in compliance with OSHA code." , " got mud in the tires.", " is making clicking sounds.", " is in bad need of a lube job." } 
	local random = math.random(#text)
	return text[random]
end


function weartear.get_pollution()
	local totalPollution = 0
	for chunk in game.surfaces.nauvis.get_chunks() do -- thanks keyboardhack for this function
    	totalPollution = totalPollution + game.surfaces.nauvis.get_pollution({chunk.x * 32, chunk.y * 32})
	end
	if totalPollution == 0 then totalPollution = 1 end -- prevent div by zero
	weartear.last_calculated_pollution = math.max(totalPollution, weartear.last_calculated_pollution)
end



function weartear.run_diagnostics_scan()
	for i=#global.scanners, 1, -1 do
		if global.scanners[i].valid then
			local scanner = global.scanners[i]
			for l=1, scanner.get_control_behavior().signals_count do
				scanner.get_control_behavior().set_signal(l, nil)
			end
			local position = scanner.position
			local force = scanner.force
			local scanner_network = scanner.surface.find_logistic_network_by_position(position, force)
			local signal_index = 1
			for j, roster_entry in pairs(global.roster) do
				if roster_entry.network == nil then roster_entry.network = {} end
				for k, network in pairs (roster_entry.network) do
					if network == scanner_network then 
						local signal_name = string.sub (roster_entry.entity.name, 5)
						local signal_ID = {type = "item", name = signal_name}
						local signal = {signal = signal_ID, countm = 1}
						scanner.get_control_behavior().set_signal(signal_index, signal)
						signal_index = signal_index + 1
						if weartear.notifications == "scanner" then 
							local loc_name1 = roster_entry.entity.localised_name[2]
							local loc_name2 = string.sub(loc_name1, 4)
							for j, player in pairs(roster_entry.entity.force.players) do
								local localized_flavor = ("flavor-text." .. roster_entry.flavor)
								player.add_custom_alert(roster_entry.entity, {type = "item", name = "scrap-metal"}, {localized_flavor, roster_entry.entity.localised_name}, true)
							end
						end
					end
				end
			end
		else
			table.remove(global.scanners, i)
		end
	end
end

function weartear.get_inserters(entity) -- get all inserters that feed given entity
	local position = entity.position
	local selection_box = entity.prototype.selection_box
	local box = {{(position.x+selection_box.left_top.x), (position.y+selection_box.left_top.y)}, {(position.x+selection_box.right_bottom.x), (position.y+selection_box.right_bottom.y)}}
	local scan_area = {{(box[1][1]-3), (box[1][2]-3)}, {(box[2][1]+3), (box[2][2]+3)}}
	local inserter_list = entity.surface.find_entities_filtered{area = scan_area, type = "inserter"}
	local x = position.x
	local y = position.y
	for i=#inserter_list, 1, -1 do
		local inserter = inserter_list[i]
		if inserter.drop_target then
			if not(((inserter.drop_target.position.x == x) and (inserter.drop_target.position.y == y))) then
				table.remove(inserter_list, i)
			end
		else
			table.remove(inserter_list, i)
		end
	end
	return inserter_list
end


function weartear.scan_for_autoreplace(player, force) -- scan for old/dilapidated equipment within logistics range.  Create womrk orders if criteria is met and tag for deconstruction
	global.replace_list = global.replace_list or {}
	for i, roster_entry in pairs(global.roster) do -- find old equipment
		local yup = false
		local replace_item 
		local entity_position
		local network
		local selected_network
		if (((roster_entry.stage == "old" and weartear.when_to_replace == "old") or roster_entry.stage == "dilapidated") and roster_entry.entity.valid) then
			local base_name = roster_entry.entity.name
			if roster_entry.stage == "old" then base_name = string.sub (base_name, 5) end
			if roster_entry.stage == "dilapidated" then base_name = string.sub (base_name, 13) end
			entity_position = roster_entry.entity.position
			local network_list = player.surface.find_logistic_networks_by_construction_area(entity_position, force)
			roster_entry.network = network_list
			for j, network in pairs(network_list) do -- iterate over all networks with construction area touching the position
				local items = game.entity_prototypes[base_name].items_to_place_this -- unfortunately this is an array with one item in it, we just need the item
				for j, part in pairs(items) do
					replace_item = part -- there's only going to be one replacement item, just grab whatever one item is in the array
				end
				local parts_available = network.get_item_count(replace_item.name)
				local parts_busy = 0
				for j, roster_entity_1 in pairs(global.replace_list) do
					if (roster_entity_1.network == network and roster_entity_1.part == replace_item) then parts_busy = parts_busy + 1 end
				end
				if parts_available > parts_busy then
					yup = true
					selected_network = network
				end
			end
			if yup == true then
				for j, replace_item in pairs(global.replace_list) do -- make sure there is not already work being done here.
					if (replace_item.position.x == entity_position.x and replace_item.position.y == entity_position.y) then 
						yup = false
					end
				end
			end
			if yup == true then
				local old_direction = roster_entry.entity.direction
				local base_name = roster_entry.entity.name
				local old_recipe				
				local old_inventory
				if roster_entry.entity.type == "assembling-machine" then -- this is to save the recipe and clean up the machine of any loose inventory (so the bots don't have to deal with it)
					if roster_entry.entity.recipe then old_recipe = roster_entry.entity.recipe.name end
					if (roster_entry.entity.is_crafting and roster_entry.entity.recipe) then -- dump in-progress items back to the inventory (which we'll then save for later)
						for k, ingredient in pairs(roster_entry.entity.recipe.ingredients) do 
							if ingredient.type == "item" then roster_entry.entity.insert({name = ingredient.name, count = ingredient.amount}) end
						end
					end
					local old_input_inventory = roster_entry.entity.get_inventory(defines.inventory.assembling_machine_input)
					if old_input_inventory then old_inventory = old_input_inventory.get_contents() roster_entry.entity.get_inventory(defines.inventory.assembling_machine_input).clear() end  -- save input inventory and destroy
					roster_entry.entity.recipe = nil 
				end
				local old_module_inventory = roster_entry.entity.get_module_inventory()
				local old_modules
				if old_module_inventory then old_modules = old_module_inventory.get_contents() roster_entry.entity.get_module_inventory().clear() end  -- save module inventory and destroy them
				if roster_entry.stage == "old" then base_name = string.sub (base_name, 5) end
				if roster_entry.stage == "dilapidated" then base_name = string.sub (base_name, 13) end
				local inserter_list = weartear.get_inserters(roster_entry.entity)
				local replace_entity = {entity_name = base_name, direction = old_direction, position = entity_position, force = force, recipe = old_recipe, modules = old_modules, network = selected_network, part = replace_item, dob = game.tick, inserters = inserter_list , inventory = old_inventory}
				for m, inserter in pairs(inserter_list) do inserter.active = false end -- turn off all feeding inserters
				table.insert(global.replace_list, replace_entity)
				roster_entry.entity.order_deconstruction(force)
			end
		end
	end
end


function weartear.scan_entities() -- routine run every 6 seconds (or interval).  Runs all the main aging functions.
	if weartear.do_this_once == true then weartear.scan_everything() weartear.do_this_once = false end
	for i, force in pairs(game.forces) do 
		for j, player in pairs(force.players) do
			if weartear.when_to_replace ~= "never" then weartear.scan_for_autoreplace(player, force) end
		end
	end
	for h, player in pairs(game.players) do
		player.remove_alert{icon = {type = "item", name = "scrap-metal"}}
	end
	for i, roster_entry in pairs(global.roster) do 
		if roster_entry.entity.valid then 
			if (weartear.notifications == "on" and (roster_entry.stage == "old" or roster_entry.stage == "dilapidated")) then 
				local loc_name1 = roster_entry.entity.localised_name[2]
--				local loc_name2 = string.sub(loc_name1, 4)
				for j, player in pairs(roster_entry.entity.force.players) do
					local localized_flavor = ("flavor-text." .. roster_entry.flavor)
					player.add_custom_alert(roster_entry.entity, {type = "item", name = "scrap-metal"}, {localized_flavor, roster_entry.entity.localised_name}, true)
				end
			end
			weartear.addwear(roster_entry, i) 
		else
			table.remove(global.roster, i)
		end 
	end
	for i, energy in pairs (weartear.max_energy_production) do -- setup max energy flow for next scan
		energy[1] = energy[2]
		energy[2] = 0
	end
	weartear.clean_schedule = weartear.clean_schedule + 1
	if weartear.clean_schedule == 10 then -- every 10 scans, purge the workorders of old entries
		weartear.clean_work_orders()
		weartear.clean_schedule = 0
	end
	weartear.run_diagnostics_scan()
end

function weartear.scan_everything() -- one time scan of the whole game on load for potential roster changes in case settings were changed. Remove any inappropriate roster entries.
	for k=#global.roster, 1, -1 do -- clean out roster entries for entities not assigned to age in the config settings
		if global.roster[k].entity.valid then
			if global.roster[k].flavor == nil then global.roster[k].flavor = math.random(weartear.flavor_size) end		
			if global.roster[k].age == nil then
				global.roster[k].age = 0 -- make sure all roster entries have an age
			end
			local info = false
			local info2 = false
			local a, b
			local type = global.roster[k].entity.type
			if (type=="mining-drill" and weartear.mining_drills_age == false) then table.remove(global.roster, k) info = true end
			if (type=="assembling-machine" and weartear.crafting_machines_age == false) then table.remove(global.roster, k) info = true end
			if (type=="generator" and weartear.steam_engines_age == false) then table.remove(global.roster, k) info = true end
			if (type=="solar-panel" and weartear.solar_panels_age == false) then table.remove(global.roster, k) info = true end
			if (type=="furnace" and weartear.furnaces_age == false) then table.remove(global.roster, k) info = true end
			if info == false then
				a,b = weartear.check_block_list(global.roster[k].entity.name)
				if (a == true) then table.remove(global.roster, k) info2 = true end
			end
			if info == true then game.print ("setting changes detected: " .. type .. "s are no longer aging") end
			if info2 == true then game.print ("setting changes detected: " .. b .. "s are no longer aging") end
		else 
			 table.remove(global.roster, k)
		end
	end
	for k, surface in pairs(game.surfaces) do
		local checklist1 = surface.find_entities_filtered{type="mining-drill"} -- get all assembling machines
		local checklist2 = surface.find_entities_filtered{type="assembling-machine"} -- get all drills
		local checklist3 = surface.find_entities_filtered{type="generator"} -- get steam engines
		local checklist4 = surface.find_entities_filtered{type="solar-panel"} -- get solar panels
		local checklist5 = surface.find_entities_filtered{type="furnace"} -- get furnaces
		for l=#checklist5, 1, -1 do -- remove non-electric furnaces
			if checklist5[l].electric_drain then  -- electric drain should be nil if not an electric energy source, according to documentation.
			-- nothing
			else
--				game.print ("scanning a furnace, it's not electric so it doesn't need to be on our list")
				table.remove(checklist5, l) end
			
		end
		if weartear.mining_drills_age == true  then weartear.scan_all(checklist1) end
		if weartear.crafting_machines_age == true then weartear.scan_all(checklist2) end
		if weartear.steam_engines_age == true  then weartear.scan_all(checklist3) end
		if weartear.solar_panels_age == true then weartear.scan_all(checklist4) end
		if weartear.furnaces_age == true then weartear.scan_all(checklist5) end
	end
end

function weartear.scan_all(list) -- one time scan to see if any existing entities need to be added to the roster.
	for i,entity in ipairs(list) do
		local entity_found = false
		for _, roster_entry in pairs(global.roster) do -- retrive entity from roster
			if entity == roster_entry.entity then 
				entity_found = true
			end
		end
		if ((entity_found == false) and (weartear.check_block_list(entity.name) == false)) then -- if we didn't find this entity on the roster, add it to the roster.
				game.print ("setting changes detected:"  .. entity.name .. "s are now aging.")
				local new_stage = "new"
				if string.find(entity.name, "Aging_") then new_stage = "aging" end
				if string.find(entity.name, "Old_") then new_stage = "old" end
				if string.find(entity.name, "Dilapidated_") then new_stage = "dilapidated" end
				local new_roster_entry = {entity = entity, age =  weartear.aging_age * .25 * (math.random()-.5), stage = new_stage, last_energy = 0, flavor = math.random(weartear.flavor_size)}
				table.insert(global.roster, new_roster_entry)	
		end
	end

end

function weartear.smoke() -- create smoke on old and dilpadated entities
	if weartear.emit_smoke then
		for i, roster_entry in pairs(global.roster) do 
			if roster_entry.entity.valid then 
				if ((roster_entry.stage == "old" and math.random(30) < 5) or (roster_entry.stage == "aging" and ((weartear.old_age - roster_entry.age) < 100) and math.random(30) < 3) or roster_entry.stage == "dilapidated") then
					roster_entry.entity.surface.create_entity{name = "old_smoke", position = roster_entry.entity.position}			
				end
			else
				table.remove(global.roster, i)
			end 
		end
	end
	global.emitter_table = global.emitter_table or {}
	for i=#global.emitter_table, 1, -1 do
		if global.emitter_table[i].valid then
			local emitter = global.emitter_table[i]
			local is_end_pipe = weartear.is_end_pipe(emitter)
			if (is_end_pipe or emitter.name == "emitter") then
				for j=#emitter.fluidbox, 1, -1 do	
					local fluid = emitter.fluidbox[j]
					if fluid then
						if fluid.type == "alien-steam" then 
							if fluid.amount > 10 then 
								global.toxin_levels = (global.toxin_levels + fluid.amount - 5)
								emitter.surface.create_entity{name="wt-toxin-cloud", position=emitter.position, force="neutral"}
								if emitter.name == "pipe" then emitter.surface.create_entity{name="vented_smoke", position={emitter.position.x,emitter.position.y}, force="neutral"} end
								fluid.amount = 5 
								emitter.fluidbox[j] = fluid 
							end
						end
						local smoke = ("wt_smoke_" .. fluid.type)
						if game.entity_prototypes[smoke] then
							emitter.surface.pollute(emitter.position, fluid.amount)
							if fluid.amount > 0 then 
								if emitter.name == "pipe" then emitter.surface.create_entity{name=smoke, position={emitter.position.x,emitter.position.y}, force="neutral"} end
								fluid.amount = 0 
								emitter.fluidbox[j] = fluid 
							end
						end
					end
				end
			else
				table.remove(global.emitter_table, i)
			end
		else
			table.remove(global.emitter_table, i)
		end
	end
end



function weartear.addwear(entry, i) -- add age to the entity, then check to see if it needs to be replaced with an older version
	if entry.entity.type == "generator" then 
		local work =  weartear.check_power_statistics(entry)
		entry.age = entry.age + work *.5 
	end 
	if entry.entity.type == "solar-panel" then entry.age = entry.age + .25 end  -- fixed amount.  
	if (entry.entity.type == "assembling-machine" or entry.entity.type == "furnace") then
		entry.age = entry.age + weartear.idle_age
		if (entry.entity.crafting_progress > 0 and entry.entity.crafting_progress < 1) then 
			entry.age = entry.age + 1 - weartear.idle_age 
			if entry.entity.recipe.name == "maintenance" then
--				game.print ("maintenance program running.  Wear is being restored")
				entry.age = entry.age - 10
			end			
		end
	end
	if entry.entity.type == "mining-drill" then
		entry.age = entry.age + weartear.idle_age
		if entry.entity.mining_progress > 0.01 then 
			entry.age = entry.age + 1 - weartear.idle_age 
		end
	end
	if entry.age > weartear.aging_age and entry.stage == "new" then weartear.tear("Aging_", entry.entity, i) end
	if entry.age > weartear.old_age and entry.stage == "aging" then weartear.tear("Old_", entry.entity, i) end
	if entry.age > weartear.dilapidated_age and entry.stage == "old" then weartear.tear("Dilapidated_", entry.entity, i) end
	if entry.age > weartear.death_age and entry.stage == "dilapidated" then entry.age = -500 weartear.death(entry.entity, i) end  
end

function weartear.death(entity) -- either explode or disable the entity.  It is now dead.
	dice_roll = math.random(10)
	if dice_roll < 4 and weartear.explosive_endings == true then
		local pos = entity.position
		local old_surface = entity.surface
		local old_force = entity.force
		old_surface.create_entity{name = "explosive-rocket", force = old_force, position = pos, target = entity, speed = 10}
		if entity.valid then entity.die() end
	else
		entity.active = false 
	end
end


function weartear.tear(stage, entity, i) -- replaces the entity with an "older" version.  Use fast_replaceable if possible, or manually adjust orientation, recipe and modules if not.
	log ("Swapping aged equipment for: " .. entity.name)
	local old_module_inventory = entity.get_module_inventory()
	local old_input_inventory = entity.get_inventory(defines.inventory.assembling_machine_input)
	local old_output_inventory = entity.get_output_inventory()
	local old_modules = nil
	local old_input = nil
	local old_output = nil
	if old_module_inventory then 
		old_modules = old_module_inventory.get_contents() 
	end 
	if old_input_inventory then 
		old_input = old_input_inventory.get_contents() 
	end 
	if old_output_inventory then 
		old_output = old_output_inventory.get_contents() 
	end 
	
	local frg = entity.prototype.fast_replaceable_group
	local pos = entity.position
	local old_surface = entity.surface
	local old_force = entity.force
	local old_direction = entity.direction
	local old_name = entity.name
	local old_type = entity.type
	local old_recipe
	local old_crafting_progress
	if old_type == "assembling-machine" then 
		old_recipe = entity.recipe 
		old_crafting_progress = entity.crafting_progress
	end
	local what_is_it = string.sub(stage,1,-2)
	local huh = string.lower(what_is_it)
	local base_name = old_name
	if stage == "Old_" then base_name = string.sub (old_name, 7) end
	if stage == "Dilapidated_" then base_name = string.sub (old_name, 5) end
	local new_name = (stage .. base_name)
	if game.entity_prototypes[new_name] then -- error checking to make sure an aging version of the prototype exists
		local replacement = old_surface.create_entity{name=new_name, force = old_force, position=pos, direction = old_direction, fast_replace = true, spill = false}
		if  (string.find(new_name, "recycler") or string.find(new_name, "refurbisher")) then table.insert (global.recyclers, replacement) end
		if new_name == "scanner" then table.insert (global.scanners, replacement) end
		if not frg then -- if there is no fast replacement group setup for this prototype, we have to manually setup it's recipe, inventory and modules
			if old_type == "assembling-machine" then 
				replacement.recipe = old_recipe 
				replacement.crafting_progress = old_crafting_progress 
			end
			if old_module_inventory  then -- re-insert any lost modules
				for item, quantity in pairs(old_modules) do 
--					game.print ("adding " .. quantity .. " " .. item .. " to " .. event.created_entity.name)
					replacement.insert({name = item, count = quantity})
				end
			end
			if old_input_inventory  then -- re-insert inputs
				for item, quantity in pairs(old_input) do 
--					game.print ("adding " .. quantity .. " " .. item .. " to " .. event.created_entity.name)
					replacement.insert({name = item, count = quantity})
				end
			end
			if old_output_inventory  then -- re-insert outputs
				for item, quantity in pairs(old_output) do 
--					game.print ("adding " .. quantity .. " " .. item .. " to " .. event.created_entity.name)
					replacement.get_output_inventory().insert({name = item, count = quantity})
				end
			end
		end
		replacement.direction = old_direction -- give it good smack on the head to make sure the direction is properly set.
		if entity.valid then -- sometimes fast-replace doesnt work properly and the old entity gets stuck.  this is to make sure it gets removed.
--			game.print ("The old one is still there?") end
			entity.destroy()
		end
--		create a new roster for this aged equipment
--		game.players[1].print ("adding " .. new_name .. "to roster.")
		if stage == "Aging_" then new_stage = "aging" end
		if stage == "Old_" then new_stage = "old" end
		if stage == "Dilapidated_" then new_stage = "dilapidated" end
		local new_entity = old_surface.find_entity(new_name, pos)
		local new_age = weartear.starting_age(new_entity, new_stage)
		local new_roster_entry = {entity = new_entity , age = new_age, stage = new_stage, last_energy = 0, flavor = math.random(weartear.flavor_size)}
		table.insert(global.roster, new_roster_entry)
	else
		game.print("Error: " .. new_name .. " does not exist. See readme.txt"	)
		table.remove(global.roster, i)
	end	
end

function weartear.starting_age(entity, stage) -- assign the starting age for this entity
	local stage_factor
	if stage == "new" then stage_factor = weartear.aging_age end
	if stage == "aging" then stage_factor = weartear.old_age end 
	if stage == "old" then stage_factor = weartear.dilapidated_age end
	if stage == "dilapidated" then stage_factor = weartear.death_age end 
	local material_bonus = weartear.get_material_bonus(entity) * stage_factor
	local random_start = (stage_factor *.25 * (math.random()-.5)) -- +/- 12% of the max age
	local age = random_start - material_bonus
	return age
end

function weartear.check_age(event) -- send this machine's age information to the chat log.	
	local player = game.players[event.player_index]
	local entity = player.selected
	if entity then 
		local position = entity.position
		for i, entry in pairs(global.roster) do
			if entry.entity.valid then
				if (entry.entity.position.x == position.x and entry.entity.position.y == position.y) then
					local units = " minutes."
					if entry.stage == "new" then
						local material_bonus = weartear.get_material_bonus(entity) * weartear.old_age
						local life = ((weartear.aging_age + weartear.old_age - entry.age + material_bonus) * weartear.interval / 3600) 
						if entity.type == "generator" then life = life * 2 end
						if entity.type == "solar-panel" then life = life * 4 end
						if life < 2 then units = " seconds." life = life * 60 end
						life = math.floor(life)
						player.print("Remaining useful life: " .. life .. units)
					end
					if entry.stage == "aging" then
						local life = ((weartear.old_age - entry.age) * weartear.interval / 3600) 
						if entity.type == "generator" then life = life * 2 end
						if entity.type == "solar-panel" then life = life * 4 end
						if life < 2 then units = " seconds." life = life * 60 end
						life = math.floor(life)
						player.print("Remaining useful life: " .. life .. units)
					end
					if entry.stage == "old" then
						local life = ((weartear.dilapidated_age - entry.age) * weartear.interval / 3600) 
						if entity.type == "generator" then life = life * 2 end
						if entity.type == "solar-panel" then life = life * 4 end
						if life < 2 then units = " seconds." life = life * 60 end
						life = math.floor(life)
						player.print("Useful life exceeded.  Becomes dilapidated in: " .. life  .. units)
					end
					if entry.stage == "dilapidated" then
						player.print("This machine is worthless.")
					end
				end
			end
		end
	end
end

function weartear.get_material_bonus(entity)
	local bonus = 0
	if entity.valid == true then
		for _, recipe in pairs(entity.force.recipes) do
			for _, product in pairs(recipe.products) do
				if product.name == entity.name then
					local ingredients = recipe.ingredients
					for j, item in pairs(ingredients) do
						for k, match in pairs(weartear.material_bonuses) do
							if string.find(item.name, match.name) then 
								bonus =math.max(bonus, match.bonus) 
							end
						end
					end
				end
			end
		end
		for _, match in pairs(weartear.special_bonuses) do
			if entity.name == match.name then
				bonus =math.max(bonus, match.bonus)
			end
		end
	end
	return bonus
end

			
function weartear.check_power_statistics(entry) -- guestimate how hard the generator is working
	local generator = entry.entity
	local last_energy = 0
	if entry.last_energy then last_energy = entry.last_energy end -- need the error check for older versions
	local x = generator.position.x
	local y = generator.position.y
	local work = weartear.idle_age
	local name = generator.name
	local output = 0
	if not weartear.max_energy_production[name] then  weartear.max_energy_production[name] = {1, 0} end -- initialize this table if it doesn't exist
	poles = generator.surface.find_entities_filtered{area = {{(x-5), (y-5)}, {(x+5), (y+5)}}, type = "electric-pole"}
	if poles[1] then
		output = ((poles[1].electric_network_statistics.get_output_count(name) - last_energy ) / 6 )
		entry.last_energy = poles[1].electric_network_statistics.get_output_count(name)
		if not entry.entity.fluidbox[1] then output = 0 end -- If there's no steam, it's not producing anything. 
		work = output / weartear.max_energy_production[name][1]
		if work ~= work then work = weartear.idle_age end
		if work <  weartear.idle_age then work = weartear.idle_age end
		if work > 1 then work = 1 end
	end
--	game.print("Generator Scan:  Name: " .. name .. "output: " .. output .. " work: " .. work .. " max energy output: " .. weartear.max_energy_production[name][1])
	if output > 0 then weartear.max_energy_production[name][2] = weartear.max_energy_production[name][2] + (generator.electric_output_flow_limit * 56.25) end -- setup max energy for next scan, only if producing energy.
	return work
end
	
	
function weartear.clean_work_orders() -- purge work orders of old and unfilled entries
	for i=#global.replace_list, 1, -1 do 
		local wo_age = (game.tick - global.replace_list[i].dob)/60/60 -- age of workorder in minutes
--		game.print("item " .. i .. " is " .. wo_age .. " minutes old...")
		if wo_age > 5 then -- anything older then 5 minutes is purged
--			game.print("found old crap, cleaning it...")
			table.remove(global.replace_list, i)
		end
	end
end
		
function weartear.check_block_list(name) -- check to see if entity is on the user-defined block list
	local result = false
	for i, list in ipairs(weartear.do_not_age) do
		if list == name then result = true end
	end
	return result, name
end


	

function weartear.onRobot_Mined(event) -- runs encoded age function, then places a ghost at the appropriate location.
	weartear.onPreMine(event) -- do the other pre-mine stuff first.
	local scan_it = false
	if (event.entity.type == "mining-drill" and weartear.mining_drills_age == true) then scan_it = true end
	if (event.entity.type == "assembling-machine" and weartear.crafting_machines_age == true) then scan_it = true end
	if (event.entity.type == "generator" and weartear.steam_engines_age == true) then scan_it = true end
	if (event.entity.type == "solar-panel" and weartear.solar_panels_age == true) then scan_it = true end
	if ((event.entity.type == "furnace" and weartear.furnaces_age == true)and event.entity.electric_drain) then scan_it = true end
	
	if scan_it == true then
		local position = event.entity.position
	
		for i, replace_item in pairs(global.replace_list) do -- search for a ghost to build at this location
			if (replace_item.position.x == position.x and replace_item.position.y == position.y) then
				if event.robot.surface.can_place_entity{name = "entity-ghost", inner_name = replace_item.entity_name, position = replace_item.position} then			
					event.robot.surface.create_entity{name = "entity-ghost", inner_name = replace_item.entity_name, force = replace_item.force, position=replace_item.position, direction = replace_item.direction, recipe = replace_item.recipe, expires = false}
				end
			end
		end
	end
end

function weartear.onRobot_Built(event) -- adjusts orientation, recipe and modules.  Closes work order. Notifies player of bot replacement. Then runs onPlayerBuilt function.
	local position = event.created_entity.position	
	for i, replace_item in pairs(global.replace_list) do -- search if this was on the replace list, we have more work to do on it
		if (replace_item.position.x == position.x and replace_item.position.y == position.y) then
			if event.created_entity.type == "assembling-machine" then 
				event.created_entity.recipe = replace_item.recipe 
				if replace_item.inventory then
					for item, quantity in pairs(replace_item.inventory) do
						event.created_entity.insert({name = item, count = quantity})
					end
				end
			end
			if replace_item.modules then -- check to see if the old entity had modules in it
				for item, quantity in pairs(replace_item.modules) do
					event.created_entity.insert({name = item, count = quantity})
				end
			end
			event.created_entity.direction = replace_item.direction -- make sure it's facing the right direction
			loc_name1 = string.gsub(event.created_entity.name, "_", "  ")
			loc_name2 = string.gsub(loc_name1, "-", "  ")
			loc_name3 = string.gsub(loc_name2, " (%a)", string.upper)
			if replace_item.inserters then
				for j, inserter in pairs(replace_item.inserters) do 
					if inserter.valid then inserter.active = true end
				end -- turn inserters back on
			end
								
			table.remove(global.replace_list, i) -- we are now finished with this replacement item, remove from the list.
		end
	end
	weartear.onPlayer_Built(event)
end


function weartear.onPlayer_Built(event) -- adds entity to the roster,if health is < max, look it up on the table, assign age and restores max health
	global.item_age_table = global.item_age_table or {}
	local add_it = false  -- we need to decide if this entity should be added to the roster
	if (event.created_entity.type == "mining-drill" and weartear.mining_drills_age == true) then add_it = true end
	if (event.created_entity.type == "assembling-machine" and weartear.crafting_machines_age == true) then add_it = true end
	if (event.created_entity.type == "generator" and weartear.steam_engines_age == true) then add_it = true end
	if (event.created_entity.type == "solar-panel" and weartear.solar_panels_age == true) then add_it = true end
	if ((event.created_entity.type == "furnace" and weartear.furnaces_age == true) and event.created_entity.electric_drain) then add_it = true end
	if weartear.check_block_list(event.created_entity.name) == true then add_it = false end
	if add_it == true then
		local flavor_text = math.random(weartear.flavor_size)
		local new_stage = "new"
		local new_age = weartear.starting_age(event.created_entity, new_stage)
		if string.find(event.created_entity.name, "Aging_") then new_stage = "aging" end
		if string.find(event.created_entity.name, "Old_") then new_stage = "old" end
		if string.find(event.created_entity.name, "Dilapidated_") then new_stage = "dilapidated" end
		if event.created_entity.health < event.created_entity.prototype.max_health then -- look for it on the item list, last item with matching name gets used.
			for i=#global.item_age_table, 1, -1 do
				if global.item_age_table[i].name == event.created_entity.name then
					new_age = global.item_age_table[i].age
--					game.print("This " .. global.item_age_table[i].name .. " is getting assigned an age of " .. new_age)
					table.remove(global.item_age_table, i)
					break
				end
			end
			if event.created_entity.health > event.created_entity.prototype.max_health * .97 then
				event.created_entity.health = event.created_entity.prototype.max_health -- if health is close to 100%, round it up to 100% on placement
			end
		end
		if new_age == nil then new_age = weartear.starting_age(event.created_entity, new_stage) end -- make sure it gets an age
		local new_roster_entry = {entity = event.created_entity, age = new_age , stage = new_stage, last_energy = 0, flavor = flavor_text}
		table.insert(global.roster, new_roster_entry)	
	end		
	if  (string.find(event.created_entity.name, "recycler") or string.find(event.created_entity.name, "refurbisher")) then table.insert (global.recyclers, event.created_entity) end
	if event.created_entity.name == "scanner" then table.insert (global.scanners, event.created_entity) end
	if event.created_entity.name == "emitter" then table.insert (global.emitter_table, event.created_entity) end
end

function weartear.onPreMine(event) -- Slightly damages entity before mining it.  Saves the item name and health in a table. 
	global.item_age_table = global.item_age_table or {}
	local damage_it = false
	if (event.entity.type == "mining-drill" and weartear.mining_drills_age == true) then damage_it = true end
	if (event.entity.type == "assembling-machine" and weartear.crafting_machines_age == true) then damage_it = true end
	if (event.entity.type == "generator" and weartear.steam_engines_age == true) then damage_it = true end
	if (event.entity.type == "solar-panel" and weartear.solar_panels_age == true) then damage_it = true end
	if ((event.entity.type == "furnace" and weartear.furnaces_age == true) and event.entity.electric_drain) then damage_it = true end
	if weartear.check_block_list(event.entity.name) == true then damage_it = false end
	if weartear.track_items == false then damage_it = false end
	if damage_it == true then
		local health = event.entity.health
		local max_health = event.entity.prototype.max_health
		local position = event.entity.position
		local stage = "new"
		local name = event.entity.name
		local age = 0 -- initialize in case someone removes an item with no age.
		for i, entry in pairs(global.roster) do -- find this entity in the roster
			if entry.entity.valid then
				if (entry.entity.position.x == position.x and entry.entity.position.y == position.y) then 
					age = entry.age
--					game.print ("The " .. name .. " picked up age is " .. age)
				end
			end
		end
		if health == max_health then health = health*.99 end
		event.entity.health = health 
		table.insert(global.item_age_table, {name = name, age = age})
	end
end

function weartear.scan_recyclers()
	for i=#global.recyclers, 1, -1 do -- clean out roster entries for entities not assigned to age in the config settings -- scan recyclers
		if global.recyclers[i].valid then
			local recycler = global.recyclers[i]
--			game.print ("Recycler found " .. i)
			if ((not recycler.is_crafting()) and recycler.get_output_inventory().is_empty()) then
				weartear.set_recycler_recipe(recycler)
			end
		else
			table.remove(global.recyclers, i)
--			game.print ("Recycle not found, removing " .. i)
		end
	end
end


function weartear.set_recycler_recipe(recycler)
	local x = recycler.position.x
	local y = recycler.position.y
	local entities
	local contents = nil
	entities = recycler.surface.find_entities_filtered{area = {{(x-2.5), (y-2.5)}, {(x+2.5), (y+2.5)}}, type = "inserter"}
	for i, inserter in pairs(entities) do
		if inserter.drop_target then
			if ((inserter.drop_target.position.x == x) and (inserter.drop_target.position.y == y)) then
				if inserter.pickup_target then
					if inserter.pickup_target.type == "transport-belt" then
						for index=1, 2 do
							if inserter.pickup_target.get_transport_line(index) then
								local beltcontents = inserter.pickup_target.get_transport_line(index).get_contents()
								for thing, count in pairs(beltcontents) do
									if string.sub(thing, 1, 4) == "Old_" then contents = beltcontents end
								end
							end
						end
					end
					if inserter.pickup_target.get_inventory(defines.inventory.item_main) then
						contents = inserter.pickup_target.get_inventory(defines.inventory.item_main).get_contents()
					end
					if contents then
						for thing, count in pairs(contents) do
							if string.sub(thing, 1, 4) == "Old_" then
								local old_inventory = 0
								if string.find(recycler.name, "recycler") then r_name = ("Recycle " .. thing) end
								if string.find(recycler.name, "refurbisher") then r_name = ("Refurbish " .. thing) end
								if recycler.get_inventory(defines.inventory.assembling_machine_input) then
									old_inventory = recycler.get_inventory(defines.inventory.assembling_machine_input).get_contents()["spare_parts"]
									recycler.clear_items_inside()
								end
								recycler.recipe = recycler.force.recipes[r_name]
								if not(old_inventory == nil) then 
									if old_inventory > 0 then recycler.insert({name="spare_parts", count=old_inventory}) end
								end
							end
						end
					end
				end
			end
		end
	end	
end

function weartear.set_suppression_levels()
	if weartear.last_calculated_pollution == 0 then weartear.get_pollution() end
	local delta = 0
	if global.critical_research_progress then -- when critical research > 10%, bugs start resisting the toxin based on progress - evolution
		if global.critical_research_progress > .1 then
			delta = (global.critical_research_progress - game.forces["enemy"].evolution_factor)
		end
	end
	if delta < 0 then delta = 0 end
	global.evolution_suppression = global.toxin_levels / (weartear.last_calculated_pollution * (1+delta*4)) 
	if global.evolution_suppression > 1.2 then global.toxin_levels = global.toxin_levels * (1-.01*(global.evolution_suppression - 1.2)^2) end
end


function weartear.is_end_pipe(entity)
	local result = false
	local neighbours = entity.neighbours
	if neighbours then 
		local the_real_size = 0
--		game.print("size of this thing is " .. #neighbours)
		for i, neighbour in pairs(neighbours) do
--			game.print("examining neighbor#"..i)
			if neighbour then
--				game.print("neighbor#" .. i .. " is real")
				the_real_size = the_real_size + 1
				if neighbour.valid then
--					game.print("neighbor#" .. i .. " is a " .. neighbour.name)
				end
			else
--				game.print("neighbor#" .. i .. " was nil.")
			end
		end
		if the_real_size == 1 then result = true end
	end
	return result
end


function weartear.toggle_boiler_status(current_entity, new_status)
	current_entity.surface.create_entity{name="vent_sound", position=current_entity.position, force="neutral"}
--	local fluidbox = current_entity.fluidbox
--	local old_fluidbox = {}
--	for i=#fluidbox, 1, -1 do
--		old_fluid = fluidbox[i]
--		if old_fluid then
--			table.insert(old_fluidbox, old_fluid)
--		end
--	end
	local fuel
	local fuel_inv = current_entity.get_fuel_inventory()
	if fuel_inv then fuel = fuel_inv.get_contents() end
	local new_emitter = current_entity.surface.create_entity{name=new_status, position=current_entity.position, force=current_entity.force, direction=current_entity.direction} 
	if new_status == "emitter" then table.insert (global.emitter_table, new_emitter) end
	current_entity.destroy()
	if fuel then 
		for item, quantity in pairs(fuel) do
			new_emitter.insert({name = item, count = quantity})
		end
	end
--	weartear.refill_tick = {game.tick+3, old_fluidbox, new_emitter}
end 

function weartear.refill_fluid(entity, old_fluidbox) 
	local new_fluidbox = entity.fluidbox
	for i=#new_fluidbox, 1, -1 do
		new_fluid = new_fluidbox[i]
		if new_fluid then
			if new_fluid.type then
				for j=#old_fluidbox, 1, -1 do
					old_fluid = old_fluidbox[j]
					if old_fluid then
						if new_fluid.type == old_fluid.type then
							new_fluid.amount = old_fluid.amount
							new_fluid.temperature = old_fluid.temperature
							new_fluidbox[i] = new_fluid
						end
					end
				end
			end
		end
	end
end


function weartear.toggle_vent(event)
	local player = game.players[event.player_index]
	local entity = player.selected
	local status
	if entity then
		local name = entity.name
		if name == "pipe" then 
			local is_end_pipe = weartear.is_end_pipe(entity)
			if is_end_pipe then
				entity.surface.create_entity{name="vent_sound", position=entity.position, force="neutral"}
				local its_on_the_list = false
				for i=#global.emitter_table, 1, -1 do
					local list_position = global.emitter_table[i].position
					if (list_position.x == entity.position.x and list_position.y == entity.position.y) then
						table.remove(global.emitter_table, i) 
						its_on_the_list = true
					end
				end
				if its_on_the_list == false then table.insert (global.emitter_table, entity)  end 
			end
		end
		if name == "emitter" then weartear.toggle_boiler_status(entity, "emitter_closed") end
		if name == "emitter_closed" then weartear.toggle_boiler_status(entity, "emitter") end
	end
end
	
function weartear.onTick(event) -- set up intervals and run routines
--	if event.tick == 1 then
--	game.players[1].cheat_mode = true
--	game.players[1].force.research_all_technologies()
--	end	
	if (event.tick % 15 == 0) then 
		weartear.onBigTick() 
		weartear.smoke()
	end
--	if event.tick == weartear.refill_tick[1] then 
--		weartear.refill_fluid(weartear.refill_tick[3], weartear.refill_tick[2])
--	end
end


function weartear.onBigTick()
	global.big_tick = global.big_tick + 1
	if (global.big_tick % 12 == 0) then 
		weartear.scan_recyclers()
		weartear.set_suppression_levels()
	end
	if (global.big_tick % 30 == 0) then weartear.scan_entities() end
	if (global.big_tick % 3000 == 0) then weartear.get_pollution() end	
end


function weartear.onInit() --Set up the global tables that are persisted between loads
	global = global or {}
	global.roster = global.roster or {}
	global.replace_list = global.replace_list or {}
	global.item_age_table = {}
	global.recyclers = global.recyclers or {}
	global.scanners = global.scanners or {}
	global.emitter_table = global.emitter_table or {}
	global.toxin_levels = global.toxin_levels or 0
	global.evolution_suppression = global.evolution_suppression or 0
	global.big_tick = global.big_tick or 1
end

function weartear.onModSettingsChange(event)
	weartear.interval = .1
	weartear.aging_age = settings.global.aging_age.value / weartear.interval 
	weartear.old_age = settings.global.old_age.value / weartear.interval 
	weartear.dilapidated_age = settings.global.dilapidated_age.value / weartear.interval 
	weartear.death_age = settings.global.death_age.value / weartear.interval 
	weartear.interval = weartear.interval * 3600
	weartear.idle_age = settings.global.idle_age.value
	weartear.notifications = settings.global.notifications.value
	weartear.emit_smoke = settings.global.smoke.value
	weartear.mining_drills_age = settings.global.mining_drills_age.value
	weartear.crafting_machines_age = settings.global.crafting_machines_age.value
	weartear.steam_engines_age = settings.global.steam_engines_age.value
	weartear.solar_panels_age = settings.global.solar_panels_age.value
	weartear.furnaces_age = settings.global.furnaces_age.value
	weartear.explosive_endings = settings.global.explosive_endings.value
	weartear.when_to_replace = settings.global.when_to_replace.value
	weartear.do_this_once = true
end

